"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .cardpaymentmethodcreate import (
    CardPaymentMethodCreate,
    CardPaymentMethodCreateTypedDict,
)
from .checkoutsessionpaymentmethodcreate import (
    CheckoutSessionPaymentMethodCreate,
    CheckoutSessionPaymentMethodCreateTypedDict,
)
from .plaidpaymentmethodcreate import (
    PlaidPaymentMethodCreate,
    PlaidPaymentMethodCreateTypedDict,
)
from .redirectpaymentmethodcreate import (
    RedirectPaymentMethodCreate,
    RedirectPaymentMethodCreateTypedDict,
)
from gr4vy.types import BaseModel
from gr4vy.utils import FieldMetadata, HeaderMetadata, RequestMetadata
import pydantic
from typing import Optional, Union
from typing_extensions import Annotated, NotRequired, TypeAliasType, TypedDict


class CreatePaymentMethodGlobalsTypedDict(TypedDict):
    merchant_account_id: NotRequired[str]
    r"""The ID of the merchant account to use for this request."""


class CreatePaymentMethodGlobals(BaseModel):
    merchant_account_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-gr4vy-merchant-account-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of the merchant account to use for this request."""


BodyTypedDict = TypeAliasType(
    "BodyTypedDict",
    Union[
        CheckoutSessionPaymentMethodCreateTypedDict,
        RedirectPaymentMethodCreateTypedDict,
        PlaidPaymentMethodCreateTypedDict,
        CardPaymentMethodCreateTypedDict,
    ],
)


Body = TypeAliasType(
    "Body",
    Union[
        CheckoutSessionPaymentMethodCreate,
        RedirectPaymentMethodCreate,
        PlaidPaymentMethodCreate,
        CardPaymentMethodCreate,
    ],
)


class CreatePaymentMethodRequestTypedDict(TypedDict):
    request_body: BodyTypedDict
    merchant_account_id: NotRequired[str]
    r"""The ID of the merchant account to use for this request."""


class CreatePaymentMethodRequest(BaseModel):
    request_body: Annotated[
        Body, FieldMetadata(request=RequestMetadata(media_type="application/json"))
    ]

    merchant_account_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-gr4vy-merchant-account-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of the merchant account to use for this request."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .approvaltarget import ApprovalTarget
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class PaymentServiceConfigurationTypedDict(TypedDict):
    approval_ui_target: ApprovalTarget
    approval_ui_height: str
    r"""Height of the approval interface in either pixels or view height (vh)."""
    approval_ui_width: str
    r"""Width of the approval interface in either pixels or view width (vw)."""
    cart_items_limit: int
    r"""The maximum number of cart items supported by this connector before we will truncate the list."""
    cart_items_required: bool
    r"""Defines if cart items are required by this connector."""
    cart_items_should_match_amount: bool
    r"""Defines if the cart items sum value should match the transaction amount."""


class PaymentServiceConfiguration(BaseModel):
    approval_ui_target: ApprovalTarget

    approval_ui_height: str
    r"""Height of the approval interface in either pixels or view height (vh)."""

    approval_ui_width: str
    r"""Width of the approval interface in either pixels or view width (vw)."""

    cart_items_limit: int
    r"""The maximum number of cart items supported by this connector before we will truncate the list."""

    cart_items_required: bool
    r"""Defines if cart items are required by this connector."""

    cart_items_should_match_amount: bool
    r"""Defines if the cart items sum value should match the transaction amount."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from datetime import datetime
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class MerchantProfileSchemeSummaryTypedDict(TypedDict):
    merchant_acquirer_bin: str
    r"""Acquirer BIN to use when calling 3DS through this scheme."""
    merchant_acquirer_id: str
    r"""Merchant ID to use when calling 3DS through this scheme."""
    merchant_name: str
    merchant_country_code: str
    r"""The merchant's ISO 3166-1 numeric country code."""
    merchant_category_code: str
    r"""Merchant category code to use when calling 3DS through this scheme."""
    merchant_url: str
    r"""URL to send when calling 3DS through this scheme."""
    created_at: datetime
    r"""The date and time when this profile was first created in our system."""


class MerchantProfileSchemeSummary(BaseModel):
    merchant_acquirer_bin: str
    r"""Acquirer BIN to use when calling 3DS through this scheme."""

    merchant_acquirer_id: str
    r"""Merchant ID to use when calling 3DS through this scheme."""

    merchant_name: str

    merchant_country_code: str
    r"""The merchant's ISO 3166-1 numeric country code."""

    merchant_category_code: str
    r"""Merchant category code to use when calling 3DS through this scheme."""

    merchant_url: str
    r"""URL to send when calling 3DS through this scheme."""

    created_at: datetime
    r"""The date and time when this profile was first created in our system."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .transactionaction import TransactionAction, TransactionActionTypedDict
from gr4vy.types import BaseModel
from typing import List
from typing_extensions import TypedDict


class TransactionActionsTypedDict(TypedDict):
    items: List[TransactionActionTypedDict]
    r"""The list of actions triggered for a transaction."""


class TransactionActions(BaseModel):
    items: List[TransactionAction]
    r"""The list of actions triggered for a transaction."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .flow import Flow
from .flowaction import FlowAction
from datetime import datetime
from gr4vy.types import BaseModel
from gr4vy.utils import validate_const
import pydantic
from pydantic.functional_validators import AfterValidator
from typing import Any, Dict, Literal, Optional
from typing_extensions import Annotated, TypedDict


class TransactionActionTypedDict(TypedDict):
    id: FlowAction
    flow: Flow
    rule_id: str
    r"""The ID of the rule that triggered this action."""
    created_at: datetime
    r"""The date this action was created at."""
    outcome: Dict[str, Any]
    r"""The outcome of the action."""
    type: Literal["action"]
    r"""Always `action`."""


class TransactionAction(BaseModel):
    id: FlowAction

    flow: Flow

    rule_id: str
    r"""The ID of the rule that triggered this action."""

    created_at: datetime
    r"""The date this action was created at."""

    outcome: Dict[str, Any]
    r"""The outcome of the action."""

    TYPE: Annotated[
        Annotated[
            Optional[Literal["action"]], AfterValidator(validate_const("action"))
        ],
        pydantic.Field(alias="type"),
    ] = "action"
    r"""Always `action`."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import UnrecognizedStr
from typing import Literal, Union


FlowAction = Union[
    Literal[
        "select-payment-options",
        "route-transaction",
        "decline-early",
        "skip-3ds",
    ],
    UnrecognizedStr,
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .billingdetails import BillingDetails, BillingDetailsTypedDict
from .shippingdetailscreate import ShippingDetailsCreate, ShippingDetailsCreateTypedDict
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing_extensions import NotRequired, TypedDict


class GuestBuyerTypedDict(TypedDict):
    display_name: NotRequired[Nullable[str]]
    r"""The display name for the buyer."""
    external_identifier: NotRequired[Nullable[str]]
    r"""The merchant identifier for this buyer."""
    billing_details: NotRequired[Nullable[BillingDetailsTypedDict]]
    r"""The billing name, address, email, and other fields for this buyer."""
    account_number: NotRequired[Nullable[str]]
    r"""The buyer account number"""
    shipping_details: NotRequired[Nullable[ShippingDetailsCreateTypedDict]]
    r"""The optional shipping details for this buyer."""


class GuestBuyer(BaseModel):
    display_name: OptionalNullable[str] = UNSET
    r"""The display name for the buyer."""

    external_identifier: OptionalNullable[str] = UNSET
    r"""The merchant identifier for this buyer."""

    billing_details: OptionalNullable[BillingDetails] = UNSET
    r"""The billing name, address, email, and other fields for this buyer."""

    account_number: OptionalNullable[str] = UNSET
    r"""The buyer account number"""

    shipping_details: OptionalNullable[ShippingDetailsCreate] = UNSET
    r"""The optional shipping details for this buyer."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "display_name",
            "external_identifier",
            "billing_details",
            "account_number",
            "shipping_details",
        ]
        nullable_fields = [
            "display_name",
            "external_identifier",
            "billing_details",
            "account_number",
            "shipping_details",
        ]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from .httpclient import AsyncHttpClient, ClientOwner, HttpClient, close_clients
from .sdkconfiguration import SDKConfiguration
from .utils.logger import Logger, get_default_logger
from .utils.retries import RetryConfig
from gr4vy import models, utils
from gr4vy._hooks import SDKHooks
from gr4vy.models import internal
from gr4vy.types import OptionalNullable, UNSET
import httpx
import importlib
import sys
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING, Union, cast
import weakref

if TYPE_CHECKING:
    from gr4vy.account_updater import AccountUpdater
    from gr4vy.audit_logs import AuditLogs
    from gr4vy.buyers_sdk import BuyersSDK
    from gr4vy.card_scheme_definitions_sdk import CardSchemeDefinitionsSDK
    from gr4vy.checkout_sessions import CheckoutSessions
    from gr4vy.digital_wallets_sdk import DigitalWalletsSDK
    from gr4vy.gift_cards_sdk import GiftCardsSDK
    from gr4vy.merchant_accounts_sdk import MerchantAccountsSDK
    from gr4vy.payment_links_sdk import PaymentLinksSDK
    from gr4vy.payment_methods_sdk import PaymentMethodsSDK
    from gr4vy.payment_options_sdk import PaymentOptionsSDK
    from gr4vy.payment_service_definitions_sdk import PaymentServiceDefinitionsSDK
    from gr4vy.payment_services_sdk import PaymentServicesSDK
    from gr4vy.payouts import Payouts
    from gr4vy.refunds_sdk import RefundsSDK
    from gr4vy.report_executions_sdk import ReportExecutionsSDK
    from gr4vy.reports_sdk import ReportsSDK
    from gr4vy.transactions import Transactions


class Gr4vy(BaseSDK):
    r"""Gr4vy: The Gr4vy API."""

    account_updater: "AccountUpdater"
    buyers: "BuyersSDK"
    payment_methods: "PaymentMethodsSDK"
    gift_cards: "GiftCardsSDK"
    card_scheme_definitions: "CardSchemeDefinitionsSDK"
    digital_wallets: "DigitalWalletsSDK"
    transactions: "Transactions"
    refunds: "RefundsSDK"
    payment_options: "PaymentOptionsSDK"
    payment_service_definitions: "PaymentServiceDefinitionsSDK"
    payment_services: "PaymentServicesSDK"
    audit_logs: "AuditLogs"
    reports: "ReportsSDK"
    report_executions: "ReportExecutionsSDK"
    checkout_sessions: "CheckoutSessions"
    merchant_accounts: "MerchantAccountsSDK"
    payouts: "Payouts"
    payment_links: "PaymentLinksSDK"
    _sub_sdk_map = {
        "account_updater": ("gr4vy.account_updater", "AccountUpdater"),
        "buyers": ("gr4vy.buyers_sdk", "BuyersSDK"),
        "payment_methods": ("gr4vy.payment_methods_sdk", "PaymentMethodsSDK"),
        "gift_cards": ("gr4vy.gift_cards_sdk", "GiftCardsSDK"),
        "card_scheme_definitions": (
            "gr4vy.card_scheme_definitions_sdk",
            "CardSchemeDefinitionsSDK",
        ),
        "digital_wallets": ("gr4vy.digital_wallets_sdk", "DigitalWalletsSDK"),
        "transactions": ("gr4vy.transactions", "Transactions"),
        "refunds": ("gr4vy.refunds_sdk", "RefundsSDK"),
        "payment_options": ("gr4vy.payment_options_sdk", "PaymentOptionsSDK"),
        "payment_service_definitions": (
            "gr4vy.payment_service_definitions_sdk",
            "PaymentServiceDefinitionsSDK",
        ),
        "payment_services": ("gr4vy.payment_services_sdk", "PaymentServicesSDK"),
        "audit_logs": ("gr4vy.audit_logs", "AuditLogs"),
        "reports": ("gr4vy.reports_sdk", "ReportsSDK"),
        "report_executions": ("gr4vy.report_executions_sdk", "ReportExecutionsSDK"),
        "checkout_sessions": ("gr4vy.checkout_sessions", "CheckoutSessions"),
        "merchant_accounts": ("gr4vy.merchant_accounts_sdk", "MerchantAccountsSDK"),
        "payouts": ("gr4vy.payouts", "Payouts"),
        "payment_links": ("gr4vy.payment_links_sdk", "PaymentLinksSDK"),
    }

    def __init__(
        self,
        bearer_auth: Optional[Union[Optional[str], Callable[[], Optional[str]]]] = None,
        merchant_account_id: Optional[str] = None,
        id: Optional[str] = None,
        server: Optional[str] = None,
        server_url: Optional[str] = None,
        url_params: Optional[Dict[str, str]] = None,
        client: Optional[HttpClient] = None,
        async_client: Optional[AsyncHttpClient] = None,
        retry_config: OptionalNullable[RetryConfig] = UNSET,
        timeout_ms: Optional[int] = None,
        debug_logger: Optional[Logger] = None,
    ) -> None:
        r"""Instantiates the SDK configuring it with the provided parameters.

        :param bearer_auth: The bearer_auth required for authentication
        :param merchant_account_id: Configures the merchant_account_id parameter for all supported operations
        :param id: Allows setting the id variable for url substitution
        :param server: The server by name to use for all methods
        :param server_url: The server URL to use for all methods
        :param url_params: Parameters to optionally template the server URL with
        :param client: The HTTP client to use for all synchronous methods
        :param async_client: The Async HTTP client to use for all asynchronous methods
        :param retry_config: The retry configuration to use for all supported methods
        :param timeout_ms: Optional request timeout applied to each operation in milliseconds
        """
        client_supplied = True
        if client is None:
            client = httpx.Client(follow_redirects=True)
            client_supplied = False

        assert issubclass(
            type(client), HttpClient
        ), "The provided client must implement the HttpClient protocol."

        async_client_supplied = True
        if async_client is None:
            async_client = httpx.AsyncClient(follow_redirects=True)
            async_client_supplied = False

        if debug_logger is None:
            debug_logger = get_default_logger()

        assert issubclass(
            type(async_client), AsyncHttpClient
        ), "The provided async_client must implement the AsyncHttpClient protocol."

        security: Any = None
        if callable(bearer_auth):
            # pylint: disable=unnecessary-lambda-assignment
            security = lambda: models.Security(bearer_auth=bearer_auth())
        else:
            security = models.Security(bearer_auth=bearer_auth)

        if server_url is not None:
            if url_params is not None:
                server_url = utils.template_url(server_url, url_params)

        server_defaults: Dict[str, Dict[str, str]] = {
            "sandbox": {
                "id": id or "example",
            },
            "production": {
                "id": id or "example",
            },
        }

        _globals = internal.Globals(
            merchant_account_id=utils.get_global_from_env(
                merchant_account_id, "GR4VY_MERCHANT_ACCOUNT_ID", str
            ),
        )

        BaseSDK.__init__(
            self,
            SDKConfiguration(
                client=client,
                client_supplied=client_supplied,
                async_client=async_client,
                async_client_supplied=async_client_supplied,
                globals=_globals,
                security=security,
                server_url=server_url,
                server=server,
                server_defaults=server_defaults,
                retry_config=retry_config,
                timeout_ms=timeout_ms,
                debug_logger=debug_logger,
            ),
            parent_ref=self,
        )

        hooks = SDKHooks()

        # pylint: disable=protected-access
        self.sdk_configuration.__dict__["_hooks"] = hooks

        current_server_url, *_ = self.sdk_configuration.get_server_details()
        server_url, self.sdk_configuration.client = hooks.sdk_init(
            current_server_url, client
        )
        if current_server_url != server_url:
            self.sdk_configuration.server_url = server_url

        weakref.finalize(
            self,
            close_clients,
            cast(ClientOwner, self.sdk_configuration),
            self.sdk_configuration.client,
            self.sdk_configuration.client_supplied,
            self.sdk_configuration.async_client,
            self.sdk_configuration.async_client_supplied,
        )

    def dynamic_import(self, modname, retries=3):
        for attempt in range(retries):
            try:
                return importlib.import_module(modname)
            except KeyError:
                # Clear any half-initialized module and retry
                sys.modules.pop(modname, None)
                if attempt == retries - 1:
                    break
        raise KeyError(f"Failed to import module '{modname}' after {retries} attempts")

    def __getattr__(self, name: str):
        if name in self._sub_sdk_map:
            module_path, class_name = self._sub_sdk_map[name]
            try:
                module = self.dynamic_import(module_path)
                klass = getattr(module, class_name)
                instance = klass(self.sdk_configuration, parent_ref=self)
                setattr(self, name, instance)
                return instance
            except ImportError as e:
                raise AttributeError(
                    f"Failed to import module {module_path} for attribute {name}: {e}"
                ) from e
            except AttributeError as e:
                raise AttributeError(
                    f"Failed to find class {class_name} in module {module_path} for attribute {name}: {e}"
                ) from e

        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def __dir__(self):
        default_attrs = list(super().__dir__())
        lazy_attrs = list(self._sub_sdk_map.keys())
        return sorted(list(set(default_attrs + lazy_attrs)))

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if (
            self.sdk_configuration.client is not None
            and not self.sdk_configuration.client_supplied
        ):
            self.sdk_configuration.client.close()
        self.sdk_configuration.client = None

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if (
            self.sdk_configuration.async_client is not None
            and not self.sdk_configuration.async_client_supplied
        ):
            await self.sdk_configuration.async_client.aclose()
        self.sdk_configuration.async_client = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class PaypalShippingOptionsItemAmountTypedDict(TypedDict):
    currency_code: str
    r"""The three-character ISO currency code."""
    value: str
    r"""The amount value, which might include a decimal portion."""


class PaypalShippingOptionsItemAmount(BaseModel):
    currency_code: str
    r"""The three-character ISO currency code."""

    value: str
    r"""The amount value, which might include a decimal portion."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

import asyncio
import random
import time
from datetime import datetime
from email.utils import parsedate_to_datetime
from typing import List, Optional

import httpx


class BackoffStrategy:
    initial_interval: int
    max_interval: int
    exponent: float
    max_elapsed_time: int

    def __init__(
        self,
        initial_interval: int,
        max_interval: int,
        exponent: float,
        max_elapsed_time: int,
    ):
        self.initial_interval = initial_interval
        self.max_interval = max_interval
        self.exponent = exponent
        self.max_elapsed_time = max_elapsed_time


class RetryConfig:
    strategy: str
    backoff: BackoffStrategy
    retry_connection_errors: bool

    def __init__(
        self, strategy: str, backoff: BackoffStrategy, retry_connection_errors: bool
    ):
        self.strategy = strategy
        self.backoff = backoff
        self.retry_connection_errors = retry_connection_errors


class Retries:
    config: RetryConfig
    status_codes: List[str]

    def __init__(self, config: RetryConfig, status_codes: List[str]):
        self.config = config
        self.status_codes = status_codes


class TemporaryError(Exception):
    response: httpx.Response
    retry_after: Optional[int]

    def __init__(self, response: httpx.Response):
        self.response = response
        self.retry_after = _parse_retry_after_header(response)


class PermanentError(Exception):
    inner: Exception

    def __init__(self, inner: Exception):
        self.inner = inner


def _parse_retry_after_header(response: httpx.Response) -> Optional[int]:
    """Parse Retry-After header from response.

    Returns:
        Retry interval in milliseconds, or None if header is missing or invalid.
    """
    retry_after_header = response.headers.get("retry-after")
    if not retry_after_header:
        return None

    try:
        seconds = float(retry_after_header)
        return round(seconds * 1000)
    except ValueError:
        pass

    try:
        retry_date = parsedate_to_datetime(retry_after_header)
        delta = (retry_date - datetime.now(retry_date.tzinfo)).total_seconds()
        return round(max(0, delta) * 1000)
    except (ValueError, TypeError):
        pass

    return None


def _get_sleep_interval(
    exception: Exception,
    initial_interval: int,
    max_interval: int,
    exponent: float,
    retries: int,
) -> float:
    """Get sleep interval for retry with exponential backoff.

    Args:
        exception: The exception that triggered the retry.
        initial_interval: Initial retry interval in milliseconds.
        max_interval: Maximum retry interval in milliseconds.
        exponent: Base for exponential backoff calculation.
        retries: Current retry attempt count.

    Returns:
        Sleep interval in seconds.
    """
    if (
        isinstance(exception, TemporaryError)
        and exception.retry_after is not None
        and exception.retry_after > 0
    ):
        return exception.retry_after / 1000

    sleep = (initial_interval / 1000) * exponent**retries + random.uniform(0, 1)
    return min(sleep, max_interval / 1000)


def retry(func, retries: Retries):
    if retries.config.strategy == "backoff":

        def do_request() -> httpx.Response:
            res: httpx.Response
            try:
                res = func()

                for code in retries.status_codes:
                    if "X" in code.upper():
                        code_range = int(code[0])

                        status_major = res.status_code / 100

                        if code_range <= status_major < code_range + 1:
                            raise TemporaryError(res)
                    else:
                        parsed_code = int(code)

                        if res.status_code == parsed_code:
                            raise TemporaryError(res)
            except httpx.ConnectError as exception:
                if retries.config.retry_connection_errors:
                    raise

                raise PermanentError(exception) from exception
            except httpx.TimeoutException as exception:
                if retries.config.retry_connection_errors:
                    raise

                raise PermanentError(exception) from exception
            except TemporaryError:
                raise
            except Exception as exception:
                raise PermanentError(exception) from exception

            return res

        return retry_with_backoff(
            do_request,
            retries.config.backoff.initial_interval,
            retries.config.backoff.max_interval,
            retries.config.backoff.exponent,
            retries.config.backoff.max_elapsed_time,
        )

    return func()


async def retry_async(func, retries: Retries):
    if retries.config.strategy == "backoff":

        async def do_request() -> httpx.Response:
            res: httpx.Response
            try:
                res = await func()

                for code in retries.status_codes:
                    if "X" in code.upper():
                        code_range = int(code[0])

                        status_major = res.status_code / 100

                        if code_range <= status_major < code_range + 1:
                            raise TemporaryError(res)
                    else:
                        parsed_code = int(code)

                        if res.status_code == parsed_code:
                            raise TemporaryError(res)
            except httpx.ConnectError as exception:
                if retries.config.retry_connection_errors:
                    raise

                raise PermanentError(exception) from exception
            except httpx.TimeoutException as exception:
                if retries.config.retry_connection_errors:
                    raise

                raise PermanentError(exception) from exception
            except TemporaryError:
                raise
            except Exception as exception:
                raise PermanentError(exception) from exception

            return res

        return await retry_with_backoff_async(
            do_request,
            retries.config.backoff.initial_interval,
            retries.config.backoff.max_interval,
            retries.config.backoff.exponent,
            retries.config.backoff.max_elapsed_time,
        )

    return await func()


def retry_with_backoff(
    func,
    initial_interval=500,
    max_interval=60000,
    exponent=1.5,
    max_elapsed_time=3600000,
):
    start = round(time.time() * 1000)
    retries = 0

    while True:
        try:
            return func()
        except PermanentError as exception:
            raise exception.inner
        except Exception as exception:  # pylint: disable=broad-exception-caught
            now = round(time.time() * 1000)
            if now - start > max_elapsed_time:
                if isinstance(exception, TemporaryError):
                    return exception.response

                raise

            sleep = _get_sleep_interval(
                exception, initial_interval, max_interval, exponent, retries
            )
            time.sleep(sleep)
            retries += 1


async def retry_with_backoff_async(
    func,
    initial_interval=500,
    max_interval=60000,
    exponent=1.5,
    max_elapsed_time=3600000,
):
    start = round(time.time() * 1000)
    retries = 0

    while True:
        try:
            return await func()
        except PermanentError as exception:
            raise exception.inner
        except Exception as exception:  # pylint: disable=broad-exception-caught
            now = round(time.time() * 1000)
            if now - start > max_elapsed_time:
                if isinstance(exception, TemporaryError):
                    return exception.response

                raise

            sleep = _get_sleep_interval(
                exception, initial_interval, max_interval, exponent, retries
            )
            await asyncio.sleep(sleep)
            retries += 1



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .giftcardredemption import GiftCardRedemption, GiftCardRedemptionTypedDict
from .giftcardservice import GiftCardService, GiftCardServiceTypedDict
from .instrumenttype import InstrumentType
from .method import Method
from .shippingdetails import ShippingDetails, ShippingDetailsTypedDict
from .transactionbuyer import TransactionBuyer, TransactionBuyerTypedDict
from .transactionintent import TransactionIntent
from .transactionpaymentmethod import (
    TransactionPaymentMethod,
    TransactionPaymentMethodTypedDict,
)
from .transactionpaymentservice import (
    TransactionPaymentService,
    TransactionPaymentServiceTypedDict,
)
from .transactionstatus import TransactionStatus
from datetime import datetime
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from gr4vy.utils import validate_const
import pydantic
from pydantic import model_serializer
from pydantic.functional_validators import AfterValidator
from typing import List, Literal, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class TransactionSummaryTypedDict(TypedDict):
    r"""A transaction, summarised"""

    id: str
    r"""The ID for the transaction."""
    reconciliation_id: str
    r"""The base62 encoded transaction ID. This represents a shorter version of this transaction's `id` which is sent to payment services, anti-fraud services, and other connectors. You can use this ID to reconcile a payment service's transaction against our system. This ID is sent instead of the transaction ID because not all services support 36 digit identifiers."""
    merchant_account_id: str
    r"""The ID of the merchant account this transaction belongs to."""
    currency: str
    r"""The currency code for this transaction."""
    amount: int
    r"""The total amount for this transaction across all funding sources including gift cards."""
    status: TransactionStatus
    authorized_amount: int
    r"""The amount for this transaction that has been authorized for the `payment_method`. This can be less than the `amount` if gift cards were used."""
    captured_amount: int
    r"""The total amount captured for this transaction, in the smallest currency unit (for example, cents or pence). This can be the full value of the `authorized_amount` or less."""
    refunded_amount: int
    r"""The total amount refunded for this transaction, in the smallest currency unit (for example, cents or pence). This can be the full value of the `captured_amount` or less."""
    settled_amount: int
    r"""The net amount settled for this transaction, in the smallest currency unit (for example, cents or pence)."""
    settled: bool
    r"""Indicates whether this transaction has been settled."""
    intent: TransactionIntent
    gift_card_redemptions: List[GiftCardRedemptionTypedDict]
    r"""The gift cards redeemed for this transaction."""
    created_at: datetime
    r"""The date and time when the transaction was created, in ISO 8601 format."""
    updated_at: datetime
    r"""The date and time when the transaction was last updated, in ISO 8601 format."""
    disputed: bool
    r"""Indicates whether this transaction has been disputed."""
    type: Literal["transaction"]
    r"""Always `transaction`."""
    settled_currency: NotRequired[Nullable[str]]
    r"""The ISO 4217 currency code of this transaction's settlement."""
    country: NotRequired[Nullable[str]]
    r"""The 2-letter ISO 3166-1 alpha-2 country code for the transaction. Used to filter payment services for processing."""
    external_identifier: NotRequired[Nullable[str]]
    r"""An external identifier that can be used to match the transaction against your own records."""
    payment_method: NotRequired[Nullable[TransactionPaymentMethodTypedDict]]
    r"""The payment method used for this transaction."""
    method: NotRequired[Nullable[Method]]
    r"""The method used for the transaction."""
    instrument_type: NotRequired[Nullable[InstrumentType]]
    r"""The name of the instrument used to process the transaction."""
    error_code: NotRequired[Nullable[str]]
    r"""The standardized error code set by Gr4vy."""
    payment_service: NotRequired[Nullable[TransactionPaymentServiceTypedDict]]
    r"""The payment service used for this transaction."""
    pending_review: NotRequired[bool]
    r"""Whether a manual anti fraud review is pending with an anti fraud service."""
    buyer: NotRequired[Nullable[TransactionBuyerTypedDict]]
    r"""The buyer used for this transaction."""
    raw_response_code: NotRequired[Nullable[str]]
    r"""This is the response code received from the payment service. This can be set to any value and is not standardized across different payment services."""
    raw_response_description: NotRequired[Nullable[str]]
    r"""This is the response description received from the payment service. This can be set to any value and is not standardized across different payment services."""
    shipping_details: NotRequired[Nullable[ShippingDetailsTypedDict]]
    r"""The shipping details associated with the transaction."""
    checkout_session_id: NotRequired[Nullable[str]]
    r"""The identifier for the checkout session this transaction is associated with."""
    gift_card_service: NotRequired[Nullable[GiftCardServiceTypedDict]]
    r"""The gift card service used for this transaction."""


class TransactionSummary(BaseModel):
    r"""A transaction, summarised"""

    id: str
    r"""The ID for the transaction."""

    reconciliation_id: str
    r"""The base62 encoded transaction ID. This represents a shorter version of this transaction's `id` which is sent to payment services, anti-fraud services, and other connectors. You can use this ID to reconcile a payment service's transaction against our system. This ID is sent instead of the transaction ID because not all services support 36 digit identifiers."""

    merchant_account_id: str
    r"""The ID of the merchant account this transaction belongs to."""

    currency: str
    r"""The currency code for this transaction."""

    amount: int
    r"""The total amount for this transaction across all funding sources including gift cards."""

    status: TransactionStatus

    authorized_amount: int
    r"""The amount for this transaction that has been authorized for the `payment_method`. This can be less than the `amount` if gift cards were used."""

    captured_amount: int
    r"""The total amount captured for this transaction, in the smallest currency unit (for example, cents or pence). This can be the full value of the `authorized_amount` or less."""

    refunded_amount: int
    r"""The total amount refunded for this transaction, in the smallest currency unit (for example, cents or pence). This can be the full value of the `captured_amount` or less."""

    settled_amount: int
    r"""The net amount settled for this transaction, in the smallest currency unit (for example, cents or pence)."""

    settled: bool
    r"""Indicates whether this transaction has been settled."""

    intent: TransactionIntent

    gift_card_redemptions: List[GiftCardRedemption]
    r"""The gift cards redeemed for this transaction."""

    created_at: datetime
    r"""The date and time when the transaction was created, in ISO 8601 format."""

    updated_at: datetime
    r"""The date and time when the transaction was last updated, in ISO 8601 format."""

    disputed: bool
    r"""Indicates whether this transaction has been disputed."""

    TYPE: Annotated[
        Annotated[
            Optional[Literal["transaction"]],
            AfterValidator(validate_const("transaction")),
        ],
        pydantic.Field(alias="type"),
    ] = "transaction"
    r"""Always `transaction`."""

    settled_currency: OptionalNullable[str] = UNSET
    r"""The ISO 4217 currency code of this transaction's settlement."""

    country: OptionalNullable[str] = UNSET
    r"""The 2-letter ISO 3166-1 alpha-2 country code for the transaction. Used to filter payment services for processing."""

    external_identifier: OptionalNullable[str] = UNSET
    r"""An external identifier that can be used to match the transaction against your own records."""

    payment_method: OptionalNullable[TransactionPaymentMethod] = UNSET
    r"""The payment method used for this transaction."""

    method: OptionalNullable[Method] = UNSET
    r"""The method used for the transaction."""

    instrument_type: OptionalNullable[InstrumentType] = UNSET
    r"""The name of the instrument used to process the transaction."""

    error_code: OptionalNullable[str] = UNSET
    r"""The standardized error code set by Gr4vy."""

    payment_service: OptionalNullable[TransactionPaymentService] = UNSET
    r"""The payment service used for this transaction."""

    pending_review: Optional[bool] = False
    r"""Whether a manual anti fraud review is pending with an anti fraud service."""

    buyer: OptionalNullable[TransactionBuyer] = UNSET
    r"""The buyer used for this transaction."""

    raw_response_code: OptionalNullable[str] = UNSET
    r"""This is the response code received from the payment service. This can be set to any value and is not standardized across different payment services."""

    raw_response_description: OptionalNullable[str] = UNSET
    r"""This is the response description received from the payment service. This can be set to any value and is not standardized across different payment services."""

    shipping_details: OptionalNullable[ShippingDetails] = UNSET
    r"""The shipping details associated with the transaction."""

    checkout_session_id: OptionalNullable[str] = UNSET
    r"""The identifier for the checkout session this transaction is associated with."""

    gift_card_service: OptionalNullable[GiftCardService] = UNSET
    r"""The gift card service used for this transaction."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "type",
            "settled_currency",
            "country",
            "external_identifier",
            "payment_method",
            "method",
            "instrument_type",
            "error_code",
            "payment_service",
            "pending_review",
            "buyer",
            "raw_response_code",
            "raw_response_description",
            "shipping_details",
            "checkout_session_id",
            "gift_card_service",
        ]
        nullable_fields = [
            "settled_currency",
            "country",
            "external_identifier",
            "payment_method",
            "method",
            "instrument_type",
            "error_code",
            "payment_service",
            "buyer",
            "raw_response_code",
            "raw_response_description",
            "shipping_details",
            "checkout_session_id",
            "gift_card_service",
        ]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .paymentoptionrequest import PaymentOptionRequest, PaymentOptionRequestTypedDict
from gr4vy.types import BaseModel
from gr4vy.utils import FieldMetadata, HeaderMetadata, RequestMetadata
import pydantic
from typing import Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class ListPaymentOptionsGlobalsTypedDict(TypedDict):
    merchant_account_id: NotRequired[str]
    r"""The ID of the merchant account to use for this request."""


class ListPaymentOptionsGlobals(BaseModel):
    merchant_account_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-gr4vy-merchant-account-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of the merchant account to use for this request."""


class ListPaymentOptionsRequestTypedDict(TypedDict):
    payment_option_request: PaymentOptionRequestTypedDict
    merchant_account_id: NotRequired[str]
    r"""The ID of the merchant account to use for this request."""


class ListPaymentOptionsRequest(BaseModel):
    payment_option_request: Annotated[
        PaymentOptionRequest,
        FieldMetadata(request=RequestMetadata(media_type="application/json")),
    ]

    merchant_account_id: Annotated[
        Optional[str],
        pydantic.Field(alias="x-gr4vy-merchant-account-id"),
        FieldMetadata(header=HeaderMetadata(style="simple", explode=False)),
    ] = None
    r"""The ID of the merchant account to use for this request."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .cardscheme import CardScheme
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from gr4vy.utils import validate_const
import pydantic
from pydantic import model_serializer
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class PaymentMethodCardTypedDict(TypedDict):
    number: str
    r"""The 13-19 digit number for this card as it can be found on the front of the card."""
    expiration_date: str
    r"""The expiration date of the card, formatted `MM/YY`."""
    method: Literal["card"]
    r"""Set to `card` to use a new card."""
    card_scheme: NotRequired[Nullable[CardScheme]]
    r"""The optional card's network scheme."""
    external_identifier: NotRequired[Nullable[str]]
    r"""The merchant identifier for this card."""


class PaymentMethodCard(BaseModel):
    number: str
    r"""The 13-19 digit number for this card as it can be found on the front of the card."""

    expiration_date: str
    r"""The expiration date of the card, formatted `MM/YY`."""

    METHOD: Annotated[
        Annotated[Optional[Literal["card"]], AfterValidator(validate_const("card"))],
        pydantic.Field(alias="method"),
    ] = "card"
    r"""Set to `card` to use a new card."""

    card_scheme: OptionalNullable[CardScheme] = UNSET
    r"""The optional card's network scheme."""

    external_identifier: OptionalNullable[str] = UNSET
    r"""The merchant identifier for this card."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["method", "card_scheme", "external_identifier"]
        nullable_fields = ["card_scheme", "external_identifier"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .refundtargettype import RefundTargetType
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class TransactionRefundCreateTypedDict(TypedDict):
    amount: NotRequired[Nullable[int]]
    r"""The amount to refund, in the smallest currency unit (e.g., cents). If omitted, a full refund will be requested."""
    target_type: NotRequired[RefundTargetType]
    target_id: NotRequired[Nullable[str]]
    r"""The optional ID of the instrument to refund for. This is only required when the `target_type` is set to `gift-card-redemption`."""
    reason: NotRequired[Nullable[str]]
    r"""An optional reason to attach extra context to the refund request."""
    external_identifier: NotRequired[Nullable[str]]
    r"""An external identifier that can be used to match the refund against your own records."""


class TransactionRefundCreate(BaseModel):
    amount: OptionalNullable[int] = UNSET
    r"""The amount to refund, in the smallest currency unit (e.g., cents). If omitted, a full refund will be requested."""

    target_type: Optional[RefundTargetType] = None

    target_id: OptionalNullable[str] = UNSET
    r"""The optional ID of the instrument to refund for. This is only required when the `target_type` is set to `gift-card-redemption`."""

    reason: OptionalNullable[str] = UNSET
    r"""An optional reason to attach extra context to the refund request."""

    external_identifier: OptionalNullable[str] = UNSET
    r"""An external identifier that can be used to match the refund against your own records."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "amount",
            "target_type",
            "target_id",
            "reason",
            "external_identifier",
        ]
        nullable_fields = ["amount", "target_id", "reason", "external_identifier"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .reportexecutioncontext import (
    ReportExecutionContext,
    ReportExecutionContextTypedDict,
)
from .reportexecutionstatus import ReportExecutionStatus
from .reportsummary import ReportSummary, ReportSummaryTypedDict
from datetime import datetime
from gr4vy.types import BaseModel
from gr4vy.utils import validate_const
import pydantic
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, TypedDict


class ReportExecutionTypedDict(TypedDict):
    id: str
    r"""The unique ID for the report execution."""
    created_at: datetime
    r"""The date this report execution was created at."""
    updated_at: datetime
    r"""The date this report execution was last updated."""
    status: ReportExecutionStatus
    context: ReportExecutionContextTypedDict
    report: ReportSummaryTypedDict
    type: Literal["report-execution"]
    r"""Always `report-execution`."""


class ReportExecution(BaseModel):
    id: str
    r"""The unique ID for the report execution."""

    created_at: datetime
    r"""The date this report execution was created at."""

    updated_at: datetime
    r"""The date this report execution was last updated."""

    status: ReportExecutionStatus

    context: ReportExecutionContext

    report: ReportSummary

    TYPE: Annotated[
        Annotated[
            Optional[Literal["report-execution"]],
            AfterValidator(validate_const("report-execution")),
        ],
        pydantic.Field(alias="type"),
    ] = "report-execution"
    r"""Always `report-execution`."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .cardscheme import CardScheme
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing_extensions import NotRequired, TypedDict


class ThreeDSecureDataV1TypedDict(TypedDict):
    cavv: str
    r"""The cardholder authentication value or AAV."""
    eci: str
    r"""The ecommerce indicator for the 3DS transaction."""
    version: str
    r"""The version of 3-D Secure that was used."""
    directory_response: str
    r"""For 3-D Secure version 1, the enrolment response. For 3-D Secure version 2 and above, the transaction status from the `ARes`."""
    authentication_response: str
    r"""The response for the 3DS authentication call."""
    cavv_algorithm: str
    r"""The CAVV algorithm used."""
    xid: str
    r"""The transaction identifier."""
    scheme: NotRequired[Nullable[CardScheme]]
    r"""The scheme/brand of the card that is used for 3-D Secure."""


class ThreeDSecureDataV1(BaseModel):
    cavv: str
    r"""The cardholder authentication value or AAV."""

    eci: str
    r"""The ecommerce indicator for the 3DS transaction."""

    version: str
    r"""The version of 3-D Secure that was used."""

    directory_response: str
    r"""For 3-D Secure version 1, the enrolment response. For 3-D Secure version 2 and above, the transaction status from the `ARes`."""

    authentication_response: str
    r"""The response for the 3DS authentication call."""

    cavv_algorithm: str
    r"""The CAVV algorithm used."""

    xid: str
    r"""The transaction identifier."""

    scheme: OptionalNullable[CardScheme] = UNSET
    r"""The scheme/brand of the card that is used for 3-D Secure."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["scheme"]
        nullable_fields = ["scheme"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .method import Method
from gr4vy.types import BaseModel
from gr4vy.utils import validate_const
import pydantic
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, TypedDict


class TransactionPaymentServiceTypedDict(TypedDict):
    id: str
    r"""The ID for the payment-service."""
    payment_service_definition_id: str
    r"""The definition ID of the service used to process this payment."""
    method: Method
    display_name: str
    r"""The display name for the payment service."""
    type: Literal["payment-service"]
    r"""Always `payment-service`."""


class TransactionPaymentService(BaseModel):
    id: str
    r"""The ID for the payment-service."""

    payment_service_definition_id: str
    r"""The definition ID of the service used to process this payment."""

    method: Method

    display_name: str
    r"""The display name for the payment service."""

    TYPE: Annotated[
        Annotated[
            Optional[Literal["payment-service"]],
            AfterValidator(validate_const("payment-service")),
        ],
        pydantic.Field(alias="type"),
    ] = "payment-service"
    r"""Always `payment-service`."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .gr4vyerror import Gr4vyError
from typing import TYPE_CHECKING
from importlib import import_module
import builtins
import sys

if TYPE_CHECKING:
    from .apierror import APIError
    from .error400 import Error400, Error400Data
    from .error401 import Error401, Error401Data
    from .error403 import Error403, Error403Data
    from .error404 import Error404, Error404Data
    from .error405 import Error405, Error405Data
    from .error409 import Error409, Error409Data
    from .error425 import Error425, Error425Data
    from .error429 import Error429, Error429Data
    from .error500 import Error500, Error500Data
    from .error502 import Error502, Error502Data
    from .error504 import Error504, Error504Data
    from .httpvalidationerror import HTTPValidationError, HTTPValidationErrorData
    from .no_response_error import NoResponseError
    from .responsevalidationerror import ResponseValidationError

__all__ = [
    "APIError",
    "Error400",
    "Error400Data",
    "Error401",
    "Error401Data",
    "Error403",
    "Error403Data",
    "Error404",
    "Error404Data",
    "Error405",
    "Error405Data",
    "Error409",
    "Error409Data",
    "Error425",
    "Error425Data",
    "Error429",
    "Error429Data",
    "Error500",
    "Error500Data",
    "Error502",
    "Error502Data",
    "Error504",
    "Error504Data",
    "Gr4vyError",
    "HTTPValidationError",
    "HTTPValidationErrorData",
    "NoResponseError",
    "ResponseValidationError",
]

_dynamic_imports: dict[str, str] = {
    "APIError": ".apierror",
    "Error400": ".error400",
    "Error400Data": ".error400",
    "Error401": ".error401",
    "Error401Data": ".error401",
    "Error403": ".error403",
    "Error403Data": ".error403",
    "Error404": ".error404",
    "Error404Data": ".error404",
    "Error405": ".error405",
    "Error405Data": ".error405",
    "Error409": ".error409",
    "Error409Data": ".error409",
    "Error425": ".error425",
    "Error425Data": ".error425",
    "Error429": ".error429",
    "Error429Data": ".error429",
    "Error500": ".error500",
    "Error500Data": ".error500",
    "Error502": ".error502",
    "Error502Data": ".error502",
    "Error504": ".error504",
    "Error504Data": ".error504",
    "HTTPValidationError": ".httpvalidationerror",
    "HTTPValidationErrorData": ".httpvalidationerror",
    "NoResponseError": ".no_response_error",
    "ResponseValidationError": ".responsevalidationerror",
}


def dynamic_import(modname, retries=3):
    for attempt in range(retries):
        try:
            return import_module(modname, __package__)
        except KeyError:
            # Clear any half-initialized module and retry
            sys.modules.pop(modname, None)
            if attempt == retries - 1:
                break
    raise KeyError(f"Failed to import module '{modname}' after {retries} attempts")


def __getattr__(attr_name: str) -> object:
    module_name = _dynamic_imports.get(attr_name)
    if module_name is None:
        raise AttributeError(
            f"No {attr_name} found in _dynamic_imports for module name -> {__name__} "
        )

    try:
        module = dynamic_import(module_name)
        result = getattr(module, attr_name)
        return result
    except ImportError as e:
        raise ImportError(
            f"Failed to import {attr_name} from {module_name}: {e}"
        ) from e
    except AttributeError as e:
        raise AttributeError(
            f"Failed to get {attr_name} from {module_name}: {e}"
        ) from e


def __dir__():
    lazy_attrs = builtins.list(_dynamic_imports.keys())
    return builtins.sorted(lazy_attrs)



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class MonatoSpeiOptionsTypedDict(TypedDict):
    approval_url: str
    r"""Approval URL that will receive a charge payment method reference."""


class MonatoSpeiOptions(BaseModel):
    approval_url: str
    r"""Approval URL that will receive a charge payment method reference."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .capturestatus import CaptureStatus
from .transaction import Transaction, TransactionTypedDict
from gr4vy.types import BaseModel, Nullable, UNSET_SENTINEL
from gr4vy.utils import validate_const
import pydantic
from pydantic import model_serializer
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, TypedDict


class TransactionCaptureTypedDict(TypedDict):
    status: CaptureStatus
    code: Nullable[str]
    r"""The standardized error code set by Gr4vy."""
    raw_response_code: Nullable[str]
    r"""This is the response code received from the payment service. This can be set to any value and is not standardized across different payment services."""
    raw_response_description: Nullable[str]
    r"""This is the response description received from the payment service. This can be set to any value and is not standardized across different payment services."""
    transaction: TransactionTypedDict
    r"""A full transaction resource."""
    type: Literal["transaction-capture"]
    r"""Always `transaction-capture`."""


class TransactionCapture(BaseModel):
    status: CaptureStatus

    code: Nullable[str]
    r"""The standardized error code set by Gr4vy."""

    raw_response_code: Nullable[str]
    r"""This is the response code received from the payment service. This can be set to any value and is not standardized across different payment services."""

    raw_response_description: Nullable[str]
    r"""This is the response description received from the payment service. This can be set to any value and is not standardized across different payment services."""

    transaction: Transaction
    r"""A full transaction resource."""

    TYPE: Annotated[
        Annotated[
            Optional[Literal["transaction-capture"]],
            AfterValidator(validate_const("transaction-capture")),
        ],
        pydantic.Field(alias="type"),
    ] = "transaction-capture"
    r"""Always `transaction-capture`."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["type"]
        nullable_fields = ["code", "raw_response_code", "raw_response_description"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .forterantifraudoptionscartitembasicitemdata import (
    ForterAntiFraudOptionsCartItemBasicItemData,
    ForterAntiFraudOptionsCartItemBasicItemDataTypedDict,
)
from .forterantifraudoptionscartitembeneficiary import (
    ForterAntiFraudOptionsCartItemBeneficiary,
    ForterAntiFraudOptionsCartItemBeneficiaryTypedDict,
)
from .forterantifraudoptionscartitemdeliverydetails import (
    ForterAntiFraudOptionsCartItemDeliveryDetails,
    ForterAntiFraudOptionsCartItemDeliveryDetailsTypedDict,
)
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing import List
from typing_extensions import NotRequired, TypedDict


class ForterAntiFraudOptionsCartItemTypedDict(TypedDict):
    basic_item_data: NotRequired[
        Nullable[ForterAntiFraudOptionsCartItemBasicItemDataTypedDict]
    ]
    r"""Basic information about the cart item."""
    delivery_details: NotRequired[
        Nullable[ForterAntiFraudOptionsCartItemDeliveryDetailsTypedDict]
    ]
    r"""Details about how the item will be delivered."""
    beneficiaries: NotRequired[
        Nullable[List[ForterAntiFraudOptionsCartItemBeneficiaryTypedDict]]
    ]
    r"""List of beneficiaries who will receive this item."""


class ForterAntiFraudOptionsCartItem(BaseModel):
    basic_item_data: OptionalNullable[ForterAntiFraudOptionsCartItemBasicItemData] = (
        UNSET
    )
    r"""Basic information about the cart item."""

    delivery_details: OptionalNullable[
        ForterAntiFraudOptionsCartItemDeliveryDetails
    ] = UNSET
    r"""Details about how the item will be delivered."""

    beneficiaries: OptionalNullable[List[ForterAntiFraudOptionsCartItemBeneficiary]] = (
        UNSET
    )
    r"""List of beneficiaries who will receive this item."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["basic_item_data", "delivery_details", "beneficiaries"]
        nullable_fields = ["basic_item_data", "delivery_details", "beneficiaries"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from pydantic import model_serializer
from typing_extensions import NotRequired, TypedDict


class StatementDescriptorTypedDict(TypedDict):
    r"""Information to show the user on their payments statement"""

    name: NotRequired[Nullable[str]]
    r"""Reflects your doing business as (DBA) name."""
    description: NotRequired[Nullable[str]]
    r"""A short description about the purchase."""
    city: NotRequired[Nullable[str]]
    r"""The merchant's city to be displayed in a statement descriptor."""
    country: NotRequired[Nullable[str]]
    r"""The 2-letter ISO country code of the merchant to be displayed in a statement descriptor."""
    phone_number: NotRequired[Nullable[str]]
    r"""The value in the phone number field of a customer's statement which should be formatted according to the E164 number standard."""
    url: NotRequired[Nullable[str]]
    r"""The merchant's URL to be displayed in a statement descriptor."""
    postal_code: NotRequired[Nullable[str]]
    r"""The merchant's postal code or zip code."""


class StatementDescriptor(BaseModel):
    r"""Information to show the user on their payments statement"""

    name: OptionalNullable[str] = UNSET
    r"""Reflects your doing business as (DBA) name."""

    description: OptionalNullable[str] = UNSET
    r"""A short description about the purchase."""

    city: OptionalNullable[str] = UNSET
    r"""The merchant's city to be displayed in a statement descriptor."""

    country: OptionalNullable[str] = UNSET
    r"""The 2-letter ISO country code of the merchant to be displayed in a statement descriptor."""

    phone_number: OptionalNullable[str] = UNSET
    r"""The value in the phone number field of a customer's statement which should be formatted according to the E164 number standard."""

    url: OptionalNullable[str] = UNSET
    r"""The merchant's URL to be displayed in a statement descriptor."""

    postal_code: OptionalNullable[str] = UNSET
    r"""The merchant's postal code or zip code."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = [
            "name",
            "description",
            "city",
            "country",
            "phone_number",
            "url",
            "postal_code",
        ]
        nullable_fields = [
            "name",
            "description",
            "city",
            "country",
            "phone_number",
            "url",
            "postal_code",
        ]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class ClickToPaySessionTypedDict(TypedDict):
    digital_payment_application_id: str
    r"""The ID of the Click to Pay application."""
    digital_payment_application_name: str
    r"""The merchant name as configured ont he the Click to Pay wallet."""


class ClickToPaySession(BaseModel):
    digital_payment_application_id: str
    r"""The ID of the Click to Pay application."""

    digital_payment_application_name: str
    r"""The merchant name as configured ont he the Click to Pay wallet."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .giftcardserviceprovider import GiftCardServiceProvider
from gr4vy.types import BaseModel
from gr4vy.utils import validate_const
import pydantic
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, TypedDict


class GiftCardServiceTypedDict(TypedDict):
    id: str
    r"""The ID for the gift card service."""
    gift_card_service_definition_id: GiftCardServiceProvider
    display_name: str
    r"""The display name for the gift card service."""
    type: Literal["gift-card-service"]
    r"""Always `gift-card-service`."""


class GiftCardService(BaseModel):
    id: str
    r"""The ID for the gift card service."""

    gift_card_service_definition_id: GiftCardServiceProvider

    display_name: str
    r"""The display name for the gift card service."""

    TYPE: Annotated[
        Annotated[
            Optional[Literal["gift-card-service"]],
            AfterValidator(validate_const("gift-card-service")),
        ],
        pydantic.Field(alias="type"),
    ] = "gift-card-service"
    r"""Always `gift-card-service`."""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel, Nullable, OptionalNullable, UNSET, UNSET_SENTINEL
from gr4vy.utils import validate_const
import pydantic
from pydantic import model_serializer
from pydantic.functional_validators import AfterValidator
from typing import Literal, Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class TokenPaymentMethodCreateTypedDict(TypedDict):
    r"""Create a transaction with the ID (token) of a stored payment method (and an optional URL for approval)"""

    id: str
    r"""The ID for the payment method."""
    method: Literal["id"]
    r"""Always `id`."""
    security_code: NotRequired[Nullable[str]]
    r"""The 3 or 4 digit security code often found on the card. This often referred to as the CVV or CVD."""
    redirect_url: NotRequired[Nullable[str]]
    r"""The URL to redirect a user back to after they approve the transaction in the browser."""


class TokenPaymentMethodCreate(BaseModel):
    r"""Create a transaction with the ID (token) of a stored payment method (and an optional URL for approval)"""

    id: str
    r"""The ID for the payment method."""

    METHOD: Annotated[
        Annotated[Optional[Literal["id"]], AfterValidator(validate_const("id"))],
        pydantic.Field(alias="method"),
    ] = "id"
    r"""Always `id`."""

    security_code: OptionalNullable[str] = UNSET
    r"""The 3 or 4 digit security code often found on the card. This often referred to as the CVV or CVD."""

    redirect_url: OptionalNullable[str] = UNSET
    r"""The URL to redirect a user back to after they approve the transaction in the browser."""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = ["method", "security_code", "redirect_url"]
        nullable_fields = ["security_code", "redirect_url"]
        null_default_fields = []

        serialized = handler(self)

        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)
            serialized.pop(k, None)

            optional_nullable = k in optional_fields and k in nullable_fields
            is_set = (
                self.__pydantic_fields_set__.intersection({n})
                or k in null_default_fields
            )  # pylint: disable=no-member

            if val is not None and val != UNSET_SENTINEL:
                m[k] = val
            elif val != UNSET_SENTINEL and (
                not k in optional_fields or (optional_nullable and is_set)
            ):
                m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from gr4vy.types import BaseModel
from typing_extensions import TypedDict


class GiftCardStoredRequestTypedDict(TypedDict):
    r"""The ID of a stored gift card to fetch a balance for."""

    id: str
    r"""The ID of the field to fetch a balance for."""


class GiftCardStoredRequest(BaseModel):
    r"""The ID of a stored gift card to fetch a balance for."""

    id: str
    r"""The ID of the field to fetch a balance for."""
